#include "uhttps.h"
#include "http_parser.h"
#include "cfg_parser.h"
#include "file_cache.h"
#include "logger.h"
#include "usersdb.h"
#include <sys/stat.h>
#include <dirent.h>
#include <time.h>


extern struct uhttp_configuration current_config;
extern struct users_db registered_usersdb;


// PRIVATE FUNCTIONS DEFINITIONS

static void _get_local_pathname( char* pathname, const char* url ) // converte una url in un pathname locale secondo le impostazioni attuali del server
{
	strcpy(pathname,current_config.root_pathname); // il primo elemento del path è la root directory
	if(strlen(url)>1)
	strcat(pathname,url);
}

static boolean _fexist(char* filename) // controlla se un file esiste senza aprirlo. ALTERNATIVES: http://stackoverflow.com/questions/230062/whats-the-best-way-to-check-if-a-file-exists-in-c-cross-platform
{
	struct stat buf;
	int i = stat ( filename, &buf );
	if ( i == 0 ) return true; // file found
	else return false;
}

static boolean _isdir(char* pathname) // ritorna "true" se pathname è una directory, o false altrimenti
{
	struct stat buf;
	stat( pathname, &buf );
	if( S_ISDIR( buf.st_mode ) ) return true;
	else return false;
}

#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	// Mingw has no "inet_ntop", so we provide an implementation using "WSAAddressToString"
	// code from http://ted.mielczarek.org/code/wget-modified/trunk/wget/src/mswindows.c
	#define xzero(x) memset (&(x), '\0', sizeof (x))
	/* An inet_ntop implementation that uses WSAAddressToString.
   Prototype complies with POSIX 1003.1-2004.  This is only used under
   IPv6 because Wget prints IPv4 addresses using inet_ntoa.  */
	const char* inet_ntop (int af, const void *src, char *dst, socklen_t cnt)
	{
	  /* struct sockaddr can't accomodate struct sockaddr_in6. */
	  union {
		struct sockaddr_in6 sin6;
		struct sockaddr_in sin;
	  } sa;
	  DWORD dstlen = cnt;
	  size_t srcsize;

	  xzero (sa);
	  switch (af)
		{
		case AF_INET:
		  sa.sin.sin_family = AF_INET;
		  sa.sin.sin_addr = *(struct in_addr *) src;
		  srcsize = sizeof (sa.sin);
		  break;
		case AF_INET6:
		  sa.sin6.sin6_family = AF_INET6;
		  sa.sin6.sin6_addr = *(struct in6_addr *) src;
		  srcsize = sizeof (sa.sin6);
		  break;
		default:
		  abort ();
		}

	  if (WSAAddressToString ((struct sockaddr *) &sa, srcsize, NULL, dst, &dstlen) != 0)
		{
		  errno = WSAGetLastError();
		  return NULL;
		}
	  return (const char *) dst;
	}
#endif

/**
 * INPUT: "client_connected_socket" socket connesso
 * OUTPUT: "client_ip_address_str" stringa contenente l'indirizzo ip del peer connesso con il socket
**/
static void _read_client_ip_address( socket_t client_connected_socket, char* client_ip_address_str )
{
	// code from http://beej.us/guide/bgnet/output/html/multipage/getpeernameman.html
	char ipstr[INET6_ADDRSTRLEN];
	int port;
	struct sockaddr_storage addr;
	socklen_t len = sizeof(addr);
	getpeername( client_connected_socket, &addr, &len );

	if (addr.ss_family == AF_INET) {
		struct sockaddr_in *s = (struct sockaddr_in *)&addr;
		port = ntohs(s->sin_port);
		inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr); // thread-safe (POSIX only)
	} else { // AF_INET6
		struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
		port = ntohs(s->sin6_port);
		inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr); // thread-safe (POSIX only)
	}
	strcpy(client_ip_address_str, ipstr);
}

// come la precedente, ma invoca "getsockname" invece di "getpeername"
static void _read_server_ip_address( socket_t client_connected_socket, char* server_ip_address_str )
{
	char ipstr[INET6_ADDRSTRLEN];
	int port;
	struct sockaddr_storage addr;
	socklen_t len = sizeof(addr);
	getsockname( client_connected_socket, &addr, &len );

	if (addr.ss_family == AF_INET) {
		struct sockaddr_in *s = (struct sockaddr_in *)&addr;
		port = ntohs(s->sin_port);
		inet_ntop(AF_INET, &s->sin_addr, ipstr, sizeof ipstr); // thread-safe (POSIX only)
		strcpy(server_ip_address_str, ipstr);
	} else { // AF_INET6
		struct sockaddr_in6 *s = (struct sockaddr_in6 *)&addr;
		port = ntohs(s->sin6_port);
		inet_ntop(AF_INET6, &s->sin6_addr, ipstr, sizeof ipstr); // thread-safe (POSIX only)
		// aggiunge le parentesi quadre attorno all'indirizzo IPv6
			strcpy(server_ip_address_str, "[");
		strcat(server_ip_address_str, ipstr);
			strcat(server_ip_address_str, "]");
	}
	if( port != 80 ) // se le porta non è quella di default, la aggiunge all'indirizzo
	{
		strcat(server_ip_address_str, ":");
		char port_str[10] = {0};
			sprintf( port_str, "%d", port );
		strcat(server_ip_address_str, port_str);
	}
}

static void _close_socket( int socket_fd )
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		shutdown( socket_fd, 2 );
		closesocket( socket_fd );
	#else
		shutdown( socket_fd, SHUT_RDWR );
		close( socket_fd );
	#endif
}

/* Read "n" bytes from a socket. */
// codice tratto dal libro "Stevens, Rago - Advanced Programming in the UNIX Environment (2e) (Addison Wesley) (2005)" (con alcune modifiche)
// ALTERNATIVA: recv flag "MSG_WAITALL" -> nn supportato su WinXP
static ssize_t recvn(int fd, void *vptr, size_t n)
{
	size_t  nleft;
    ssize_t nread;
    void   *ptr;

    ptr = vptr;
    nleft = n;
    while (nleft > 0) {
		errno=0;
        if ( (nread = recv(fd, ptr, nleft, 0)) < 0) {
            if (errno == EINTR)
                continue; //nread = 0;      /* and call read() again */
            else
                return (-1);
        } else if (nread == 0)
            break;              /* EOF */

        nleft -= nread;
        ptr += nread;
    }
    return (n - nleft);         /* return >= 0 */
}

/* Write "n" bytes to a descriptor */
// codice tratto dal libro "Stevens, Rago - Advanced Programming in the UNIX Environment (2e) (Addison Wesley) (2005)" (con alcune modifiche)
static ssize_t sendn(int fd, const void *ptr, size_t n)
{
    size_t      nleft;
    ssize_t     nwritten;

    nleft = n;
    while (nleft > 0) {
		errno=0;
        if ((nwritten = send(fd, ptr, nleft, 0)) < 0) {
			if (errno == EINTR)
				continue;
            if (nleft == n) // se la prima scrittura fallisce ritorna subito "-1"
                return(-1);
            else
                break;      /* else return amount written so far */
        } else if (nwritten == 0) {
            break;
        }
		
        nleft -= nwritten;
        ptr   += nwritten;
    }
    return(n - nleft);      /* return >= 0 */
}

#define _READ_CHUNK_SIZE 4096
// ritorna il numero di bytes occupati dagli headers della request
// oppure "-1" in caso di errore di ricezione (headers incompleti).
// NOTA: lascia i dati nel buffer usando il recv flag "MSG_PEEK"
static size_t _read_request_headers_size( socket_t client_connected_socket )
{
	// PROBLEMA: the size of the incoming data is unknown
	// SOL.: FAQ 3.8 http://www.faqs.org/faqs/unix-faq/socket/
	// legge la request un chunk alla volta e rialloca il buffer se necessario
	
	char* buffer = calloc( _READ_CHUNK_SIZE, 1 );
	if(!buffer) return(-1);
	size_t buffer_current_size = _READ_CHUNK_SIZE;
	size_t r;
	int tries_count = 0;
	
	// cicla finchè non sono stati letti tutti gli headers
	// (il buffer deve contenere il marcatore della fine degli headers = 2 newlines consecutivi)
	while( !strstr(buffer, "\r\n\r\n") ) // && !strstr(buffer, "\n\n")
	{
		// dopo 10 tentativi termina e ritorna errore
		if( tries_count == 10 )
		{
			free(buffer);
			return(-1);
		}
		
		if( tries_count > 0 ) // se non è il primo tentativo...
		{
			uhttps_log_error("[error] incomplete headers received. Trying again in 1 second...");
			
			// dorme per un secondo prima di riprovare
			#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
				Sleep(1000);
			#else
				sleep(1);
			#endif
		}
			
		// legge il prossimo chunk
		errno=0;
		r = recv( client_connected_socket, buffer, buffer_current_size, MSG_PEEK );
		if( r==-1 || r==0 )
		{
			if(errno==EINTR) // se la recv è stata interrotta da un segnale la riavvia
				continue;
			// else
				//uhttps_log_error("[warn] socket error while reading request: %s", strerror(errno) );
				//strerror_r(...);
				free(buffer);
				return(-1);
		}
		
		tries_count++;
			
		if( r == buffer_current_size )
		{
			// incrementa la dimensione del buffer di 1 chunk e lo rialloca
			buffer_current_size += _READ_CHUNK_SIZE;
			buffer = realloc(buffer, buffer_current_size);
				// NOTA: realloc garantisce che il contenuto del buffer è mantenuto inalterato.
				// La sua locazione in memoria però potrebbe cambiare, quindi è necessario riscrivere il puntatore al buffer ad ogni invocazione.
			
			tries_count = 0;
		}
	
	} // end while
	
	// calcola la dimensione degli headers in bytes
		char* headers_end = strstr(buffer, "\r\n\r\n");
		size_t headers_size = headers_end + strlen("\r\n\r\n") - buffer;
	
	free(buffer);
	return headers_size;
}

// imposta un codice d'errore nella response e genera una pagina HTML ad hoc
void _set_http_response_error( struct uttps_http_response_parsed* parsed_response, const short int STATUS_CODE, const char* ADDITIONAL_MESSAGE )
{
	if(ADDITIONAL_MESSAGE==NULL) ADDITIONAL_MESSAGE = "";
		
	parsed_response->status_code = STATUS_CODE;
	
	switch( STATUS_CODE )
	{
		case(200):
			strcpy(parsed_response->status_code_description, "OK");
			break;
		case(201):
			strcpy(parsed_response->status_code_description, "Created");
			break;
		case(400):
			strcpy(parsed_response->status_code_description, "Bad Request");
			break;
		case(401):
			strcpy(parsed_response->status_code_description, "Unauthorized");
			break;
		case(403):
			strcpy(parsed_response->status_code_description, "Forbidden");
			break;
		case(404):
			strcpy(parsed_response->status_code_description, "Not Found");
			break;
		case(405):
			strcpy(parsed_response->status_code_description, "Method Not Allowed");
			break;
		case(500):
			strcpy(parsed_response->status_code_description, "Internal Server Error");
			break;
		case(501):
			strcpy(parsed_response->status_code_description, "Not Implemented");
			break;
	}

	// inserisce nel body una pagina HTML di errore
	strcpy(parsed_response->content_type, "text/html");
	parsed_response->body = calloc(100+strlen(ADDITIONAL_MESSAGE), 1); // size = dimensione codice HTML + strlen(ADDITIONAL_MESSAGE)
		sprintf(parsed_response->body, "<html>\n<body>\n<h1>%d %s</h1>\n", STATUS_CODE, (parsed_response->status_code_description)); // HTML header
		if( *ADDITIONAL_MESSAGE!=0 ) strcat(parsed_response->body, ADDITIONAL_MESSAGE);
		strcat(parsed_response->body, "\n</body>\n</html>\n"); // HTML footer
	
	parsed_response->content_length = strlen(parsed_response->body); // calcola il content-length
}


// PRIVATE FUNCTIONS FORWARD DECLARATIONS
static void _handle_http_get( const struct uttps_http_request_parsed* parsed_request, 
                                    struct uttps_http_response_parsed* response );
static void _handle_http_get_command( const struct uttps_http_request_parsed* parsed_request, 
							                struct uttps_http_response_parsed* parsed_response );
static void _handle_http_get_registry( const struct uttps_http_request_parsed* parsed_request, 
							                 struct uttps_http_response_parsed* parsed_response );
static void _handle_http_put( const struct uttps_http_request_parsed* parsed_request, 
                                    struct uttps_http_response_parsed* response );


// MODULE PUBLIC FUNCTION DEFINITIONS
void uhttps_handle_http_request( socket_t client_connected_socket )
{
	// alloca i buffer per la request e la response e li inizializza a 0
		struct uttps_http_request_parsed parsed_request; memset( &parsed_request, 0, sizeof(parsed_request));
		struct uttps_http_response_parsed parsed_response; memset( &parsed_response, 0, sizeof(parsed_response));
		char* http_response_headers_buf = calloc( 4096, 1 ); // alloca il buffer per gli headers della response (non è necessario riallocarlo?)
		if(!http_response_headers_buf)
		{
			// errore fatale: non si può leggere la request e/o non si può inviare una response. Non resta che ritornare...
			uhttps_log_error("[error] malloc error. Insufficient memory?");
			_close_socket(client_connected_socket);
			return;
		}

	// imposta gli headers sempre presenti nella response
		strcpy(parsed_response.http_version, "1.0"); // header "HTTP version"
		strcpy(parsed_response.server, "uhttps/1.0"); // header "Server"
		// header "Date"
			// parse date format according to RFC 822 (updated by RFC 1123)
			time_t now = time(NULL);
			#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
				struct tm* d = gmtime(&now); // on Windows "gmtime" is thread-safe... -> http://old.nabble.com/Porting-localtime_r-and-gmtime_r-td15282276.html
			#else
				struct tm date;
				struct tm* d = gmtime_r(&now, &date);  // ...on POSIX it is not.
			#endif
		strftime(parsed_response.date, _UHTTPS_MAX_HTTP_HEADER_LINE_LENGTH, "%a, %d %b %Y %H:%M:%S GMT", d );
		parsed_response._shared_body = false;
		_read_server_ip_address( client_connected_socket, parsed_response._server_ip_address ); // legge l'indirizzo IP del server (necessario per il directory listing)
		
	// imposta un timeout di 10 sec per la receive -> evita che un thread/processo resti bloccato a causa di un errore del client
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)		
			int timeout = 10*1000; // msec
		#else
			struct timeval timeout;
				timeout.tv_sec = 10;
				timeout.tv_usec = 0; // msec
		#endif
		setsockopt(client_connected_socket, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
	// TODO? ingrandisce il buffer in ricezione -> non necessario?
		//int nSocketBuffer = ?; // ?MB
		//setsockopt(server_socket_fd, SOL_SOCKET, SO_RCVBUF, &nSocketBuffer, sizeof(int));
	
	// legge la dimensione in bytes degli headers della request
		const size_t REQUEST_HEADERS_SIZE = _read_request_headers_size( client_connected_socket );
		if( REQUEST_HEADERS_SIZE == -1 )
		{
			uhttps_log_error("[error] error reading request headers size (request malformed?)");
			_set_http_response_error( &parsed_response, 400, NULL );
			goto send;
		}

	// alloca il buffer per leggere gli headers della request
		char* http_request_headers_buf = calloc( REQUEST_HEADERS_SIZE+10, 1 ); // alloca il buffer per gli headers della request + 1 bytes per string nul-termination + padding di "sicurezza"
		if( !http_request_headers_buf )
		{
			uhttps_log_error("[error] malloc error. Insufficient memory?");
			_set_http_response_error( &parsed_response, 500, "Please retry later." );
			goto send; // la request non può essere soddisfatta
		}
			
	// legge gli headers della request
		if( recvn( client_connected_socket, http_request_headers_buf, REQUEST_HEADERS_SIZE ) != REQUEST_HEADERS_SIZE )
		{
			uhttps_log_error("[error] incomplete headers received");
		}
		// NOTA: prosegue anche a seguito di errori (connection lost, etc...) -> logga cmq le richieste
	
	// parsa gli headers della request
		uhttps_parse_http_request_headers( &parsed_request, http_request_headers_buf );

	// controlla che la request contiene l'Initial Request Line completo
		if( *(parsed_request.method)=='\0' ||
			*(parsed_request.url)=='\0' ||
			*(parsed_request.http_version)=='\0'
		  ) {
			  uhttps_log_error("[error] bad/malformed request");
			  _set_http_response_error( &parsed_response, 400, NULL );
			  goto send;
		  }
	
	// controlla se l'header Authorization è presente -> SEMPRE RICHIESTO?
		if( *(parsed_request.username)=='\0' ||
		    *(parsed_request.password)=='\0'
		  ) {
			  strcpy(parsed_response.www_authenticate, "Basic");  // TODO? customizable realm names
			  // strcpy(parsed_response.www_authenticate, "Basic realm=\"Secure Area\"");
			  _set_http_response_error( &parsed_response, 401, "You must authenticate to access this server." );
			  goto send; // ignora la request
		  }
		
	// verifica se l'account è presente nel DB degli utenti
		if( uhttps_usersdb_login( &registered_usersdb, parsed_request.username, parsed_request.password ) == false )
		{
			// utente non registrato
			uhttps_log_error("[notice] login failed \"%s:%s\"", parsed_request.username, parsed_request.password);
			_set_http_response_error( &parsed_response, 403, "Please retry login." );
			goto send; // ignora la request
		}
	
	// se il metodo è PUT, legge anche il body della request
		if( strcmp(parsed_request.method, "PUT") == 0 )
		{
			// alloca il buffer per il body della request
			parsed_request.body = calloc( (parsed_request.content_length), 1 );
			if( parsed_request.body == NULL )
			{
				// malloc error
				uhttps_log_error("[error] malloc error. Insufficient memory?");
				_set_http_response_error( &parsed_response, 500, "Please retry the upload." );
				goto send; // la request non può essere soddisfatta
			}

			// legge il body della request
			if( recvn( client_connected_socket, parsed_request.body, parsed_request.content_length ) != parsed_request.content_length )
			{
				// error - partial body received
				uhttps_log_error("[error] bad/malformed request: Partial body received.");
				_set_http_response_error( &parsed_response, 400, "Partial body received, please retry the upload." );
				goto send; // la request viene ignorata
			}
		}
		
	// in base al metodo HTTP, invoca una funzione privata di gestione apposita
		if( strcmp(parsed_request.method, "GET") == 0 )
		{
			_handle_http_get( &parsed_request, &parsed_response );
		}
		else if( strcmp(parsed_request.method, "HEAD") == 0 )
		{
			// la response di "HEAD" è la stessa di "GET", ma senza body
			_handle_http_get( &parsed_request, &parsed_response );
			parsed_response.body = NULL;
		}
		else if( strcmp(parsed_request.method, "PUT") == 0 )
		{
			_handle_http_put( &parsed_request, &parsed_response );
		}
		else
		{
			uhttps_log_error("[error] unsupported HTTP request method \"%s\"", parsed_request.method);
			_set_http_response_error( &parsed_response, 405, "Allowed methods are: GET, PUT and HEAD." );
		}

send:
	// parsa la response
		uhttps_parse_http_response_headers( http_response_headers_buf, &parsed_response );

	// invia la response
		// effettua 2 send separate: 1 per gli headers e 1 per il body -> VANTAGGIO: evita di copiare il body nel buffer della response
		if( sendn( client_connected_socket, http_response_headers_buf, parsed_response._headers_size) != parsed_response._headers_size )
		{
			// error: partial headers sent
			uhttps_log_error("[error] socket error while sending response (headers incomplete?)");
		}
		if( parsed_response.body!=NULL && parsed_response.content_length!=0 )
		{
			// invia anche il body
			if( sendn( client_connected_socket, parsed_response.body, parsed_response.content_length) != parsed_response.content_length )
			{
				// error: partial body
				uhttps_log_error("[error] socket error while sending response (body incomplete?)");
			}
		}
		// NOTA: prosegue anche a seguito di errori (connection lost, etc...) -> logga cmq le richieste
		
	// registra nel log la richiesta e la risposta
		char client_ip_address[INET6_ADDRSTRLEN] = {0};
			_read_client_ip_address( client_connected_socket, client_ip_address );
		char http_request_line[100] = {0};
			sprintf(http_request_line, "%s %s HTTP/%s", parsed_request.method, parsed_request.url, parsed_request.http_version);
		uhttps_log_request( client_ip_address, parsed_request.username, http_request_line, parsed_response.status_code, parsed_response.content_length);
	
	// chiusura connected socket con il client
		_close_socket(client_connected_socket);
		
	// disalloca i buffers
		if(http_request_headers_buf!=NULL) free( http_request_headers_buf );
		if(http_response_headers_buf!=NULL) free( http_response_headers_buf );
		if(parsed_request.body!=NULL) free(parsed_request.body);
		if(parsed_response.body!=NULL && parsed_response._shared_body==false) free( parsed_response.body );
		
} // END OF uhttps_handle_http_request


static void _handle_http_get( const struct uttps_http_request_parsed* parsed_request, 
                                    struct uttps_http_response_parsed* parsed_response )
{	
	// se l'url inizia con "/command/..." esegue il comando ed inserisce nel body l'output
	if( strncmp( (parsed_request->url), "/command/", sizeof("/command") ) == 0 )
	{
		_handle_http_get_command( parsed_request, parsed_response );
		return;
	}
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	// se l'url inizia con "/register/..." esegue una query nel registro di Windows
	else if( strncmp( (parsed_request->url), "/register/", sizeof("/register") ) == 0 )
	{
		_handle_http_get_registry( parsed_request, parsed_response );
		return;
	}
	#endif
	
	// se non si è ritornati, l'url deve far riferimento ad un file od una directory
	// quindi controlla se esiste:
	char pathname[ _UHTTPS_MAX_PATHNAME_LENGHT ];
		_get_local_pathname( pathname, parsed_request->url );
	if( !_fexist( pathname ) )
	{
		// il file non esiste invia al client lo status code "404 Not Found"
		uhttps_log_error("[notice] Requested file does not exist: %s", pathname);
		_set_http_response_error( parsed_response, 404, "Please re-check the URL.");
		return;
	}
	// se "pathname" è una directory, stampa il suo contenuto come anchors html
	if( _isdir(pathname) )
	{
		parsed_response->status_code = 200;
		strcpy(parsed_response->status_code_description, "OK");
		strcpy(parsed_response->content_type, "text/html");
		
		// apre la directory e conta i files all'interno
		DIR* mydir = opendir( pathname );
		struct dirent* f;
		int files_count = 0;
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			while( ( f = readdir( mydir ) ) != NULL )
		#else
			struct dirent entry;
			while( ( readdir_r( mydir, &entry, &f )  == 0 ) && f )  // thread-safe (POSIX-only)
		#endif
		{
			files_count++;
		}
		rewinddir(mydir);
		
		// alloca il buffer per il body della response
		parsed_response->body = malloc( 1000 + files_count*1000 ); // 1000 bytes per i tags html + 1000 chars per ogni anchor
		if( parsed_response->body == NULL ) return;
		
		// inserisce un header html nel body
		// NOTA: usa "parsed_response->content_length" come offset per le scritture...
		parsed_response->content_length = 0;
		parsed_response->content_length +=
			sprintf(parsed_response->body, "<html>\n"
	                                       "<body>\n"
	                                       "<h1>Index of %s</h1>\n"
			                               "<pre>\n", parsed_request->url );
		
		/* SPOSTATO (serviva "client_connected_socket")
		// legge l'indirizzo ip del server (da mettere all'inizio dei link)
		char server_ip_address[INET6_ADDRSTRLEN] = {0};
			_read_server_ip_address( client_connected_socket, server_ip_address );
		*/
		
		// inserisce tanti link quanti sono i files nella directory
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			while( ( f = readdir( mydir ) ) != NULL )
		#else
			while( ( readdir_r( mydir, &entry, &f )  == 0 ) && f )  // thread-safe (POSIX-only)
		#endif
		{
			if( strcmp(f->d_name, ".") == 0 ) continue;
			if( ( strcmp(parsed_request->url,"/") == 0 ) && ( strcmp(f->d_name, "..") == 0 ) ) continue; // non consente di "risalire" oltre la root dir
			
			parsed_response->content_length +=
				sprintf((parsed_response->body)+(parsed_response->content_length), "<a href=\"http://%s%s%s%s\">%s</a>\n", parsed_response->_server_ip_address, parsed_request->url,
					(strcmp(parsed_request->url,"/")? "/" : ""), f->d_name, f->d_name );
			        // NOTA: lo slash tra "parsed_request->url" e "f->d_name" è inserito solo se "parsed_request->url" non è la root dir
			
		}
		closedir( mydir );
		
		// inserisce il footer HTML
		parsed_response->content_length +=
			sprintf((parsed_response->body)+(parsed_response->content_length), "\n</pre>\n</body>\n</html>\n");
		
		return;
	}
	// altrimenti è un file, quindi legge il suo contenuto
	
	// legge il contenuto del file dalla cache del server
	// se necessario, il file è aperto e letto dalla cache
	struct cached_file requested_file;
	switch( uhttps_read_cached_file( &requested_file, pathname, &current_config ) )
	{
		case(READ_OK):
		{
			parsed_response->body = requested_file.buffer;
			parsed_response->_shared_body = true; // disattiva la disallocazione del body dopo l'invio
			parsed_response->content_length = requested_file.length;
			parsed_response->status_code = 200;
			strcpy(parsed_response->status_code_description, "OK");
			//strcpy(parsed_response->content_type, "???"); // TODO? identificare il tipo mime
			return;
		}
				
		default:
		{
			// internal server error
			_set_http_response_error( parsed_response, 500, "Please retry later." );
		}
	}
	
} // END OF _handle_http_get


static void _handle_http_put( const struct uttps_http_request_parsed* parsed_request, 
                                    struct uttps_http_response_parsed* parsed_response )
{
	// pre-imposta un errore generico ("fallimento sicuro")
		_set_http_response_error( parsed_response, 500, "Please retry the upload." );

	char pathname[ _UHTTPS_MAX_PATHNAME_LENGHT ];
		_get_local_pathname( pathname, parsed_request->url );
	
	// contolla se il file esiste già PRIMA dell'apertura (necessario in seguito)
	boolean pathname_fexist = _fexist( pathname );
	
	// prova ad aprire il file su disco
	FILE* request_file = fopen( pathname, "wb" ); // NOTA: tronca un file esistente
	if(request_file == NULL)
	{
		// error
		// status_code = ?
		return;
	}
	
// RIMUOVERE: -> se l'apertura con troncamento è riuscita il file non era di certo lockato... -> ok solo su Windows?
/*
	// prova ad acquisire il lock esclusivo sul file (thread-safe)
	// NOTA: la chiamata non è bloccante perchè il file potrebbe restare lockato da un altro processo per un lungo periodo. Se il primo tentativo fallisce ritorna subito.
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		HANDLE request_file_handle = _get_osfhandle(_fileno(request_file)); // ottiene l'HANDLE associato al FILE stream
		if( LockFile( request_file_handle, 0, 0, parsed_request->content_length, 0 ) == 0 ) // blocca l'intero file
	#else
		if( ... != 0 )
	#endif
		{
			// lock failed:
			// response = resource in use, try again later...
			// status_code = ?
			return;
		}
*/
	
	// se il corpo non è vuoto prova a scrivere il file
	size_t written_bytes = 0;
	if( parsed_request->body != NULL )
	{
		written_bytes = fwrite((parsed_request->body), 1, (size_t)(parsed_request->content_length), request_file);
	}
	
/*
	// rilascia il lock sul file e lo chiude subito
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		UnlockFile( request_file_handle, 0, 0, parsed_request->content_length, 0 );
	#else
		funlockfile( request_file );
	#endif
*/
	fclose( request_file );
	
	// controlla l'esito della scrittura
	if( written_bytes < 0 )
	{
		// local write error
		// status_code = 500 (Internal server error) 
		return;
	}
	else if( written_bytes != (parsed_request->content_length) )
	{
		// error - partial body written
		return;
	}
	// else
	// la scrittura è andata a buon file:

	if( !pathname_fexist ) // il file è stato creato ex-novo
	{
		// response status = 201 (Created)
		uhttps_log_error("[notice] New file uploaded: %s", pathname);
		_set_http_response_error( parsed_response, 201, "File created correctly.");
		return;
	}
	else // il file esisteva già ed è stato sovrascritto/aggiornato
	{
		// response status = 200 (OK)
		uhttps_log_error("[notice] File over-written: %s", pathname);
		_set_http_response_error( parsed_response, 200, "File updated correctly.");
		return;
	}

} // END OF _handle_http_put


static void _handle_http_get_command( const struct uttps_http_request_parsed* parsed_request, 
							                struct uttps_http_response_parsed* parsed_response )
{
	// invia sempre status code 200 OK -> ALTERNATIVA: inviare "404 Not Found" se il comando non esiste...
	parsed_response->status_code = 200;
	strcpy(parsed_response->status_code_description, "OK");
	strcpy(parsed_response->content_type, "text/html");
			
	char* command = parsed_request->url + sizeof("/command");
	// no need to parse args?
	
	// pre-alloca il buffer per il body (sarà riallocato se necessario durante le letture)
	size_t buffer_size = 1000;
	parsed_response->body = malloc( 1000 );
	if( parsed_response->body == NULL ) return;
	
	// inserisce un header html nel body
	// NOTA: usa "parsed_response->content_length" come offset per le scritture...
	parsed_response->content_length = 0;
	parsed_response->content_length +=
		sprintf(parsed_response->body, "<html>\n"
									   "<body>\n"
									   "<h1>output of command &quot;%s&quot;</h1>\n"
									   "<pre>\n", command );
	
	// esegue il comando in un processo separato
	strcat( command, " 2>&1"); // ridireziona anche stderr
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		strcat( command, " 0<NUL"); // ridireziona stdin -> evita che si blocchi perchè in attesa di input (ad es. comando "date" su Windows)
		FILE* command_stdout = _popen( command, "r" );
	#else
		strcat( command, " < /dev/null");
		FILE* command_stdout = popen( command, "r" );
	#endif
	if( command_stdout==NULL )
	{
		//uhttps_log_error("[error] popen error");
		//strerror_r(...);
		return;
	}
	
	// legge il suo stdout nel body
	size_t r = fread( (parsed_response->body)+(parsed_response->content_length), 1, 512, command_stdout ); // legge il primo blocco
	parsed_response->content_length += r;
	
	 // rialloca il buffer e continua a leggere se ci sono altri dati da leggere
	while( r == 512 )
	{
		buffer_size += 512;
		parsed_response->body = realloc(parsed_response->body, buffer_size); // rialloca il buffer
		r = fread( (parsed_response->body)+(parsed_response->content_length), 1, 512, command_stdout ); // legge un altro blocco da 512
		parsed_response->content_length += r;
	}
	
	pclose(command_stdout);
	
	// chiude i tags html nel body
	parsed_response->content_length +=
		sprintf((parsed_response->body)+(parsed_response->content_length), "\n</pre>\n</body>\n</html>\n");
	
	return;
}

#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
#define _MEMORY_PAGE_SIZE 4096
void _handle_http_get_registry( const struct uttps_http_request_parsed* parsed_request, 
							          struct uttps_http_response_parsed* parsed_response )
{
	char* registry_key_addr = parsed_request->url + sizeof("/register");
	
	HKEY root_key;
	char* registry_subkey;
	if( strncmp(registry_key_addr, "HKEY_CLASSES_ROOT", strlen("HKEY_CLASSES_ROOT")) == 0 )
	{
		root_key = HKEY_CLASSES_ROOT;
		registry_subkey = registry_key_addr + strlen("HKEY_CLASSES_ROOT") + 1;
	}
	else if( strncmp(registry_key_addr, "HKEY_CURRENT_USER", strlen("HKEY_CURRENT_USER")) == 0 )
	{
		root_key = HKEY_CURRENT_USER;
		registry_subkey = registry_key_addr + strlen("HKEY_CURRENT_USER") + 1;
	}
	else if( strncmp(registry_key_addr, "HKEY_LOCAL_MACHINE", strlen("HKEY_LOCAL_MACHINE")) == 0 )
	{
		root_key = HKEY_LOCAL_MACHINE;
		registry_subkey = registry_key_addr + strlen("HKEY_LOCAL_MACHINE") + 1;
	}
	else if( strncmp(registry_key_addr, "HKEY_USERS", strlen("HKEY_USERS")) == 0 )
	{
		root_key = HKEY_USERS;
		registry_subkey = registry_key_addr + strlen("HKEY_USERS") + 1;
	}
	
	HKEY requested_key;
	if( RegOpenKeyEx( root_key, registry_subkey, 0, KEY_READ, &requested_key ) != ERROR_SUCCESS )
	{
		uhttps_log_error("[error] the requested registry key does not exist: %s", registry_key_addr);
		_set_http_response_error( parsed_response, 404, "The requested registry key does not exist." );
		return;
	}
	// else
	parsed_response->status_code = 200;
	strcpy(parsed_response->status_code_description, "OK");
	strcpy(parsed_response->content_type, "text/html");

	// pre-alloca il buffer per il body (sarà riallocato se necessario durante le letture dei valori)
	size_t current_body_size = _MEMORY_PAGE_SIZE;
	parsed_response->body = calloc( current_body_size, 1 );
	if( parsed_response->body == NULL ) return;
	
	parsed_response->content_length = 0;
	parsed_response->content_length +=
		sprintf(parsed_response->body, "<html>\n"
									   "<body>\n"
									   "<h1>Contents of the registy key &quot;%s&quot;</h1>\n"
									   "<table style=\"text-align: center; width: 100%%;\" border=\"1\" cellpadding=\"2\" cellspacing=\"2\">\n"
									   "<thead>\n<tr>\n<th>Name</th>\n<th>Type</th>\n<th>Data</th>\n</tr>\n</thead>\n<tbody>\n", registry_key_addr );
	
	// cicla finchè ci sono valori da leggere
		char* current_value_name = calloc( 16383+1, 1 );
		DWORD current_value_name_size = 16383+1;
		DWORD current_data_type = 0;
		void* current_data = NULL;
		DWORD current_data_size = 0;
		DWORD dwIndex = 0;
		const char* REGISTRY_VALUE_TYPES[] = { "REG_NONE", "REG_SZ", "REG_EXPAND_SZ", "REG_BINARY", "REG_DWORD", "REG_DWORD_LITTLE_ENDIAN", "REG_DWORD_BIG_ENDIAN", "REG_LINK", "REG_MULTI_SZ", "REG_RESOURCE_LIST", "REG_FULL_RESOURCE_DESCRIPTOR", "REG_RESOURCE_REQUIREMENTS_LIST", "REG_QWORD", "REG_QWORD_LITTLE_ENDIAN" };
	while( RegEnumValue( requested_key,
						 dwIndex,
						 current_value_name,
						 &current_value_name_size,
						 0,
						 &current_data_type,
						 NULL,
						 &current_data_size ) == ERROR_SUCCESS )
	{
//DEBUG: printf("\t\tcurrent_data_size=%d\n", current_data_size);
		if(current_data_size<=1 && (current_data_type==REG_SZ || current_data_type==REG_EXPAND_SZ))
		{
			current_data = strdup("(value not set)");
			current_data_size = strlen(current_data)+1;
		}
		else
		{
			// alloca il buffer per leggere il dato corrente
			current_data = calloc( current_data_size+5, 1 );
			
			// legge il dato (come prima, ma stavolta lpData!=NULL)
			current_value_name_size++;
			RegEnumValue( requested_key,
							 dwIndex,
							 current_value_name,
							 &current_value_name_size,
							 0,
							 &current_data_type,
							 current_data,
							 &current_data_size );
		}
		
		if(*current_value_name == '\0') // empty string
			strcpy(current_value_name, "(Default)"); // the default value
		
		// controllo body buffer overflow
			while( current_body_size <= ( (parsed_response->content_length) + current_value_name_size + current_data_size + 200 /* per codice HTML */) )
			{
				// se lo spazio nel body non è sufficiente, lo rialloca
				current_body_size += _MEMORY_PAGE_SIZE;
				parsed_response->body = realloc( parsed_response->body, current_body_size );
				if( parsed_response->body == NULL ) return;
			}
			
		// stampa nel body della response
		switch(current_data_type)
		{
			case(REG_SZ):
			case(REG_EXPAND_SZ):
			case(REG_MULTI_SZ): // TODO: stampare tutte le stringhe
			{
				// current data is a string
				//*((char*)current_data+current_data_size)-1=0; // assicura che la stringa sia nul-terminated
				parsed_response->content_length +=
				sprintf((parsed_response->body)+(parsed_response->content_length),
						"<tr>\n<td>%s</td>\n<td>%s</td>\n<td>%s</td>\n</tr>\n",
						current_value_name,
						REGISTRY_VALUE_TYPES[current_data_type],
						(char*)current_data );
				break;
			}
	
			default: // integers -> printed as hexadecimal
			{
				parsed_response->content_length +=
				sprintf((parsed_response->body)+(parsed_response->content_length),
						"<tr>\n<td>%s</td>\n<td>%s</td>\n<td>0x%.8x</td>\n</tr>\n",
						current_value_name,
						REGISTRY_VALUE_TYPES[current_data_type],
						*(int*)current_data );
			}
		} // END switch
		
		// prepara lettura prossimo valore
			dwIndex++;
			free(current_data);
			current_data = NULL;
			current_value_name_size = 16383+1;

	} // END WHILE

	free(current_value_name);
	RegCloseKey( requested_key );
	
	// chiude i tags html nel body
	parsed_response->content_length +=
		sprintf((parsed_response->body)+(parsed_response->content_length), "</tbody>\n</table>\n</body>\n</html>\n");	
	
	return;
} // END OF _handle_http_get_registry
#endif
