#include "uhttps.h"
#include "cfg_parser.h"
#include "file_cache.h"
#include <sys/stat.h>
#include <dirent.h>


// *  Non è richiesto un meccanismo di controllo che il file sia stato modificato nel periodo in cui è mantenuto in memoria. 
// * Il numero massimo di file da mantenere in memoria è fissato (ad esempio 32 file) ma modificabile tramite il file di configurazione. Se viene superato tale limite, viene eliminato il file che è da più tempo in memoria.

// la file cache è un array di "struct cached_file", la dimensione è stabilita a runtime da "current_config->max_cached_file_count"
static struct cached_file* _file_cache;
static unsigned int _file_cache_index_older; // indice del possimo file da sostistuire nella cache

// mutex per l'accesso esclusivo alla file cache (1 solo thread alla volta può accedere alla file cache) -> il lock è effettuato all'interno del modulo dalla "uhttps_read_cached_file"
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	static CRITICAL_SECTION _file_cache_mutex;
#else
	//static pthread_mutex_t* _file_cache_mutex = NULL;
	static pthread_mutex_t _file_cache_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
static void _file_cache_mutex_init()
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		InitializeCriticalSection( &_file_cache_mutex );
	#else
		//_file_cache_mutex = malloc( sizeof(pthread_mutex_t) );
		//pthread_mutex_init( _file_cache_mutex, NULL );
		// protezione deadlocks -> NON NECESSARIA?
		/*
			pthread_mutexattr_t attr;
			pthread_mutexattr_init(&attr);
			pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK); // abilita error-checking per prevenire deadlocks
		pthread_mutex_init( _file_cache_mutex, &attr );
			pthread_mutexattr_destroy(&attr);
		*/
	#endif
}
static void _file_cache_lock()
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		EnterCriticalSection( &_file_cache_mutex );
	#else
		pthread_mutex_lock( &_file_cache_mutex );
	#endif
}
static void _file_cache_unlock()
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		LeaveCriticalSection( &_file_cache_mutex );
	#else
		pthread_mutex_unlock( &_file_cache_mutex );
	#endif

}

static size_t _fsize(char* filename) // ritorna la dimensione di un file (senza aprirlo)
{
	struct stat r; memset(&r,0,sizeof(struct stat));
	stat(filename, &r);
	return(r.st_size);
}


void uhttps_init_file_cache( struct uhttp_configuration* current_config )
{
	if( !uhttps_file_cache_is_enabled(current_config) ) return;
	
	_file_cache = calloc( (current_config->max_cached_file_count), sizeof(struct cached_file) );
	if( !_file_cache )
	{
		// disattiva la file cache e ritorna
		uhttps_log_error("[error] malloc error. File cache disabled.");
		current_config->max_cached_file_count = 0;
		return;
	}
	_file_cache_index_older = 0;
	_file_cache_mutex_init();
	
	// apre la root directory e seleziona i files da mettere nella cache
	DIR* mydir = opendir( current_config->root_pathname );
	if(!mydir)
	{
		uhttps_log_error("[fatal error] unable to open root directory.");
		exit(1);
	}
	
	struct dirent* f;
	struct stat s;
	while( ( f = readdir( mydir ) ) != NULL ) // NOTA: essendo eseguito in fase di inizializzazione del server, readdir non deve essere thread-safe
	{
		if( strcmp(f->d_name, ".") == 0 ) continue;
		if( strcmp(f->d_name, "..") == 0 ) continue;
		
		// genera il pathname completo
			char pathname[ _UHTTPS_MAX_PATHNAME_LENGHT ] = {0};
			strcat(pathname,current_config->root_pathname);
			strcat(pathname,"/");
			strcat(pathname,f->d_name);
			
		// legge le proprietà del file
			memset(&s,0,sizeof(struct stat));
			stat(pathname, &s);

		// se il file è una directory la ignora
		if( S_ISDIR( s.st_mode ) )
			continue;
		// ALTERNATIVA: recurse subdirectories
		
		// se il file ha dimensione maggiore di quella specificata nel file di configurazione lo ignora
		if( s.st_size > (current_config->max_cached_file_size) )
			continue;
		// altrimenti lo legge e lo inserisce nella cache
		uhttps_read_cached_file( &(_file_cache[_file_cache_index_older]), pathname, current_config );
		
		// se la cache è piena interrompe
		if( _file_cache_index_older == (current_config->max_cached_file_count) )
			break;
	}
	// TODO: scansione ricorsiva della dirs -> escludere directories sym/hardlinkate

	uhttps_log_error("[notice] file cache initialized. %d files read.", ( _file_cache_index_older ? _file_cache_index_older : (current_config->max_cached_file_count) ) );
}

int uhttps_read_cached_file( struct cached_file* output, const char* pathname, const struct uhttp_configuration* current_config )
{
	if(!output || !pathname || !current_config) return ARG_ERROR;
	memset( output, 0, sizeof(output) );
	struct cached_file* r = output; // alias del puntatore "output"
	
	if( uhttps_file_cache_is_enabled(current_config) )
	{
		// ricerca il pathname nella cache. Se il file è cached lo ritorna subito.
		// OSS.: la ricerca è lineare
		_file_cache_lock();
			int i;
			for(i=0; i<(current_config->max_cached_file_count); i++)
			{
				if( strcmp( pathname, _file_cache[i].pathname ) == 0 )
				{
					memcpy( output, &(_file_cache[i]), sizeof(struct cached_file) );
					_file_cache_unlock();
					uhttps_log_error("[notice] file \"%s\" read from cache.", pathname );
					return READ_OK;
				}
			}
		_file_cache_unlock();
	}
	
	// altrimenti lo legge dal disco
	strcpy(r->pathname, pathname);
	r->length = _fsize(pathname);
	if( r->length == 0 ) goto file_cache_update; // se il file è vuoto non c'è bisogno di aprirlo per leggerlo...
	r->buffer = calloc( r->length, 1 ); // alloca il buffer per leggere il file
	if( r->buffer == NULL )
		return READ_ERROR;
	
	FILE* request_file = fopen( pathname, "rb" );
	if( !request_file )
	{
		free(r->buffer);
		return OPEN_ERROR;
	}
	
	// prova ad acquisire il lock esclusivo sull'intero file (thread-safe)
	// NOTA: la chiamata non è bloccante perchè il file potrebbe restare lockato da un altro processo per un lungo periodo. Se il primo tentativo fallisce ritorna subito.
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		HANDLE request_file_handle = _get_osfhandle(fileno(request_file)); // ottiene l'HANDLE associato al FILE stream
		if( LockFile( request_file_handle, 0, 0, r->length, 0 ) == 0 ) // blocca l'intero file
	#else
		int request_file_fd = fileno( request_file ); // ottiene il file descriptor associato al FILE stream
		if( lockf( request_file_fd, F_TLOCK, 0 ) != 0 ) // ALTERNATIVA: fcntl
	#endif
		{
			// lock fallito
			free(r->buffer);
			return FILE_LOCKED;
		}
	
	if( fread( r->buffer, 1, r->length, request_file ) != r->length )
	{
		free(r->buffer);
		return READ_ERROR;
	}

	// rilascia il lock sul file e lo chiude
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		UnlockFile( request_file_handle, 0, 0, r->length, 0 );
	#else
		lockf(request_file_fd, F_ULOCK, 0);
	#endif
	fclose( request_file );
	
	// aggiorna la cache
file_cache_update:
	if( uhttps_file_cache_is_enabled(current_config) )
	{
		_file_cache_lock();
			// se il file è minore delle dimensioni specificate, lo mette nella cache al posto del file più vecchio
			if( (r->length) <= (current_config->max_cached_file_size) )
			{
				free(_file_cache[_file_cache_index_older].buffer); // libera il buffer del vecchio file
				memcpy( &(_file_cache[_file_cache_index_older]), output, sizeof(struct cached_file) );
				_file_cache_index_older++;
				if( _file_cache_index_older == (current_config->max_cached_file_count) )
					_file_cache_index_older = 0; // l'indice è circolare, deve essere resettato
			}
		_file_cache_unlock();
	}
	
	return READ_OK;
}