#include "uhttps.h"
#include "client_handler.h"
#include "cfg_parser.h"
#include "logger.h"
#include "usersdb.h"
#include "file_cache.h"
#include <signal.h>


String current_config_pathname = _UHTTPS_DEFAULT_CONFIGURATION_PATHNAME;

// shared datas between threads
volatile struct uhttp_configuration current_config; // configurazioni attuali del server - acceduto in SOLA LETTURA dai threads ed in SCRITTURA alla recezione di SIGHUP
volatile struct users_db registered_usersdb; // database degli utenti - acceduto in SOLA LETTURA dai threads ed in SCRITTURA alla recezione di SIGHUP
static volatile sig_atomic_t connected_clients_count = 0; // contatore dei clients connessi - acceduto in LETTURA+SCRITTURA dai threads -> uso MUTEX semplice:
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	//NO? HANDLE connected_clients_count_mutex; -> using interlocked operations for this counter
#else
	pthread_mutex_t* connected_clients_count_mutex = NULL;
#endif

static void _init_mutexes() // NOTA: invocato solo se il server è in multithreading (test eseguito nel main)
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		// NO? CreateMutex(... -> using interlocked operations
	#else
		connected_clients_count_mutex = malloc( sizeof(pthread_mutex_t) );
		pthread_mutex_init( connected_clients_count_mutex, NULL );
		// protezione deadlocks -> NON NECESSARIA?
		/*
			pthread_mutexattr_t attr;
			pthread_mutexattr_init(&attr);
			pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK); // abilita error-checking per prevenire deadlocks
		pthread_mutex_init( connected_clients_count_mutex, &attr );
			pthread_mutexattr_destroy(&attr);
		*/
	#endif
}
//static void _free_mutexes() -> NON NECESSARIO (fatto automaticamente alla chiusura del server)

static void _increment_connected_clients_count()
{
	if( uhttps_is_multithreaded( &current_config ) ) // if using multithreading we need to acquire the lock
	{
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			InterlockedIncrement( &connected_clients_count );
		#else
			pthread_mutex_lock( connected_clients_count_mutex );
			connected_clients_count++;
			pthread_mutex_unlock( connected_clients_count_mutex );
		#endif
	}
	else // no multithreading
	{
		// just increment
		connected_clients_count++;
	}
}

static void _decrement_connected_clients_count()
{
	if( uhttps_is_multithreaded( &current_config ) ) // if using multithreading we need to acquire the lock
	{
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			InterlockedDecrement( &connected_clients_count );
		#else
			pthread_mutex_lock( connected_clients_count_mutex );
			connected_clients_count--;
			pthread_mutex_unlock( connected_clients_count_mutex );
		#endif
	}
	else // no multithreading
	{
		// just decrement
		connected_clients_count--;
	}
}

static int _read_connected_clients_count()
{
	//NO? return connected_clients_count;
	
	if( uhttps_is_multithreaded( &current_config ) ) // if using multithreading we need to acquire the lock
	{
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			return( InterlockedCompareExchange( &connected_clients_count, 0, 0 ) ); // interlocked read
		#else
			pthread_mutex_lock( connected_clients_count_mutex );
			int r = connected_clients_count;
			pthread_mutex_unlock( connected_clients_count_mutex );
			// OSS.: il valore di "connected_clients_count" potrebbe variare nel frattempo...
			return r;
		#endif
	}
	else // no multithreading
	{
		// just read
		return connected_clients_count;
	}
}

// wrapper di "uhttps_handle_http_request" per threads
// NOTA: "arg" punta a "socket_t client_connected_socket"
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
static DWORD WINAPI _handle_http_request_thread( void* arg )
#else
static void* _handle_http_request_thread( void* arg )
#endif
{
	socket_t client_connected_socket = *((socket_t*) arg);
	uhttps_handle_http_request(client_connected_socket);
	if(arg) free(arg);
	_decrement_connected_clients_count(); // decrementa il contatore dei client connessi. OSS.: in modalità multiprocesso è il signal handler di SIGCHLD a farlo...
	return NULL; // termina il thread
}

// a seconda della configurazione corrente del server,
// crea un nuovo thread o un nuovo processo per gestire una request di un client
// (invoca sempre "uhttps_handle_http_request")
static void _client_handler( socket_t client_connected_socket )
{
	if( uhttps_is_multithreaded( &current_config ) )
	{
		// gestisce la richiesta con un nuovo thread
		
		// crea una copia in memoria dinamica del socket ID per passarla al nuovo thread come argomento
		// (sarà deallocata dal thread stesso una volta terminato)
			void* arg = malloc (sizeof(client_connected_socket));
			memcpy(arg,&client_connected_socket,sizeof(client_connected_socket));
		// OSS.: void* arg = &client_connected_socket; NON VA BENE! -> a seconda di quando parte il thread, lo storage della variabile potrebbe non essere più disponibile...
		
		// crea il nuovo thread passandogli "arg" come argomento
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			//#if defined(_MT)
				// _beginthreadex requires a multithreaded C run-time library
				DWORD lpThreadId;
				if( _beginthreadex( NULL, 0, _handle_http_request_thread, arg, 0, &lpThreadId ) > 0 )
				{
					uhttps_log_error("[notice] started secondary thread");
					return;
				}
			/* NO? -> http://www.mingw.org/wiki/Use_the_thread_library
			#else
				if( CreateThread( NULL, 0, _handle_http_request_thread, &client_connected_socket, 0, NULL ) != NULL )
				{
					uhttps_log_error("[notice] started secondary thread");
					return;
				}
			#endif*/
			// else
			uhttps_log_error("[notice] error creating secondary thread");
			// se la creazione del thread fallisce gestisce la richiesta nel thread principale...
		#else
			pthread_t tid;
				pthread_attr_t attr;
				pthread_attr_init(&attr);
				pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED); // imposta l'attributo di creazione "detached/not-joinable" per risparmiare risorse
			//if( pthread_create( &tid, NULL, _handle_http_request_thread, arg ) == 0 )
			if( pthread_create( &tid, &attr, _handle_http_request_thread, arg ) == 0 )
			{
					pthread_attr_destroy(&attr);
				uhttps_log_error("[notice] started secondary thread");
				return;
			}
			// se la creazione del thread fallisce gestisce la richiesta nel thread principale...
		#endif
	}
	else if( current_config.max_processes > 1 )
	{
		// gestisce la richiesta con un nuovo processo
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			//NO? CreateProcess(... -> vari problemi...
		#else
			errno=0;
			pid_t p = fork();
			if( p == 0 )
			{
				// child
				uhttps_log_error("[notice] child %d: child process is running", getpid());
				uhttps_handle_http_request( client_connected_socket );
				exit(0);
			}
			else if( p > 0 )
			{
				// parent
				uhttps_log_error("[notice] parent: created child process %d", p);
				return;
			}
			else
			{
				uhttps_log_error("[warn] error forking: %s", strerror(errno) );
			}
			// altrimenti gestisce la richiesta nel processo principale...
		#endif
	}

	// gestisce la richiesta nel processo principale/corrente
	// OSS.: il server non può accettare nuove connessioni nel frattempo
	uhttps_handle_http_request( client_connected_socket );

}


// SIGNAL HANLDER DEFINITIONS
	// variabili usate dai signal handlers:
	static volatile socket_t server_socket; // server socket per l'ascolto delle connessioni
	static volatile boolean server_is_opened = true; // finchè è true il server resta in ascolto di connessioni
	static volatile boolean _current_config_reload_signal = false; // alla ricezione di SIGHUP viene settato a true dal relativo signal handler

// sighandler_t
void _close_server( int signo )
{
	#if !defined(WINDOWS) && !defined(WIN32) && !defined(_WIN32)
		signal( signo, _close_server ); // ristabilisce il signal handler (necessario per unreliable signals)
	#endif
	
	if( server_is_opened == true )
	{
		uhttps_log_error("[notice] first termination signal received, only one request still accepted...");
		//uhttps_log_error("[notice] first termination signal received, closing listening port...");
		server_is_opened = false;
		return;
	}
	else // esce forzatamente
	{
		uhttps_log_error("[notice] second termination signal received, exiting forcefully (%d clients were connected)...", _read_connected_clients_count());
		exit(1);
	}
}

// sighandler_t
void _close_server_imm( int signo )
{
	#if !defined(WINDOWS) && !defined(WIN32) && !defined(_WIN32)
		signal( signo, _close_server_imm ); // ristabilisce il signal handler (necessario per unreliable signals)
	#endif
	
	// chiusura immediata del server
	uhttps_log_error("[notice] termination signal received, exiting forcefully (%d clients were connected)...", _read_connected_clients_count());
	exit(1);
}

// sighandler_t
void _reload_configuration( int signo )
{
	#if !defined(WINDOWS) && !defined(WIN32) && !defined(_WIN32)
		signal(SIGHUP, _reload_configuration ); // ristabilisce il signal handler (necessario per unreliable signals)
	#endif
	
	//uhttps_log_error("[notice] SIGHUP received, server configuration will be releaded when all pending requests are done (now %d)...", _read_connected_clients_count());
	uhttps_log_error("[notice] SIGHUP received, server configuration will be releaded after next request...", _read_connected_clients_count());
	
	_current_config_reload_signal = true;
	// la re-inizializzazione del server sarà effettuata dal thread principale -> VANTAGGIO: nessuna necessità di sincronizzare la modifica a "current_config"
}

/* signal handler per SIGCHLD - RIMUOVERE?
#if !defined(WINDOWS) && !defined(WIN32) && !defined(_WIN32)
	#include <sys/wait.h>
	static void _free_child( int signo )
	{
		signal(SIGCHLD, _free_child ); // ristabilisce il signal handler (necessario per unreliable signals)
	
		// lettura exit status del figlio -> rimuove il processo zombie dal SO
		int exit_code;
		pid_t terminated_child_pid = waitpid( -1, &exit_code, WNOHANG );
		if(terminated_child_pid>0) uhttps_log_error("[notice] child %d terminated with exit code=%d", terminated_child_pid, exit_code);
	
		// decrementa il contatore dli clients connessi
		// OSS.: in modalità multiprocesso non necessita di sincronizzazione
		connected_clients_count--;
	}
#endif
*/

// console handler unico per Windows
// OSS.: passa la gestione ai signal handlers definiti precedentemente
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	BOOL WINAPI _HandlerRoutine(
		DWORD dwCtrlType   //  control signal type
	) {
		switch( dwCtrlType )
		{
			case(CTRL_CLOSE_EVENT):  // equivalente di SIGHUP
			{
				_reload_configuration( 0 );
				return TRUE;
			}
		
			case(CTRL_C_EVENT): // equivalente di SIGINT
			{
				_close_server( 0 );
				return TRUE;
				
			}
			
			case(CTRL_LOGOFF_EVENT):
			case(CTRL_SHUTDOWN_EVENT): // equivalente di SIGTERM
			{
				_close_server_imm( 0 );
				return TRUE;
			}
			
			default:
			{
				return FALSE; // ignora l'evento
			}
		}
	}
#endif

static void _register_signal_handlers()
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		SetConsoleCtrlHandler( _HandlerRoutine, TRUE );

	#else
		signal(SIGHUP, _reload_configuration);
	
		signal(SIGINT, _close_server);
	/*
			struct sigaction sa;
			sa.sa_handler = _close_server;
			sigemptyset( &sa.sa_mask );
			sa.sa_flags = SA_RESTART; // richiede esplicitamente il riavvio automatico delle syscalls interrotte
			sigaction( SIGINT, &sa, NULL );
	*/
	
		signal(SIGTERM, _close_server_imm);
		signal(SIGQUIT, _close_server_imm);
		signal(SIGABRT, _close_server_imm);
	
		signal(SIGPIPE, SIG_IGN); // evita la terminazione del programma se il client chiude la connessione durante la "send"
		signal(SIGUSR1, SIG_IGN);
		signal(SIGUSR2, SIG_IGN);
	
		/* se il server è in modalità multi-processo
	       un child non può decrementare la variabile del padre "connected_clients_count"
	       perchè modificherebbe solo la sua copia.
	    */
		if( ( current_config.max_processes > 1 ) &&
			( current_config.max_processes > current_config.max_threads )
		  ) {
				// signal(SIGCHLD, _free_child);
				struct sigaction sa;
				  sa.sa_handler = _decrement_connected_clients_count;
				  sigemptyset( &sa.sa_mask );
				  sa.sa_flags = SA_NOCLDWAIT; // non genera processi zombie
				sigaction( SIGCHLD, &sa, NULL );
		  }
	#endif
}


#if !defined(WINDOWS) && !defined(WIN32) && !defined(_WIN32)
	#include <fcntl.h>
#endif
static void _detach_main_process()
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		FreeConsole();
	#else
		// code from http://www.developerweb.net/forum/showthread.php?t=3025
		/* logica:
			1. fork -> il padre termina subito, così il figlio viene "adottato" da init
		    2. il figlio cambia il suo session/group ID
			3. fa di nuovo fork -> ???
			4. ridireziona stdin/out/err a /dev/null, in questo modo le funzioni di I/O non falliranno perchè troveranno questi descrittori chiusi
		*/
		int fd;
		switch (fork()) {
		   case 0:
			  break;
		   case -1:
			  // Error
			  fprintf(stderr, "Error demonizing (fork)! %d - %s\n", errno, strerror(errno));
			  exit(0);
			  break;
		   default:
			  _exit(0);
		}

		if (setsid() < 0) {
		  fprintf(stderr, "Error demonizing (setsid)! %d - %s\n", errno, strerror(errno));
		  exit(0);
		}
		
		switch (fork()) {
		   case 0:
			  break;
		   case -1:
			  // Error
			  fprintf(stderr, "Error demonizing (fork2)! %d - %s\n", errno, strerror(errno));
			  exit(0);
			  break;
		   default:
			  _exit(0);
		}

		//chdir("/");

		fd = open("/dev/null", O_RDONLY);
		if (fd != 0) {
		  dup2(fd, 0);
		  close(fd);
		}
		fd = open("/dev/null", O_WRONLY);
		if (fd != 1) {
		  dup2(fd, 1);
		  close(fd);
		}
		/*
		fd = open("/dev/null", O_WRONLY);
		if (fd != 2) {
		  dup2(fd, 2);
		  close(fd);
		}*/
	#endif
		freopen( "stderr.txt", "a", stderr ); // ridireziona stderr su un file
}


/* TODO? pagine d'errore customizzabili
void* 401_PAGE = NULL; // puntatore al buffer con il contenuto della pagina
void* 403_PAGE = NULL;
void* 404_PAGE = NULL;
void _read_error_pages()
{
	FILE* 401_page_file = fopen( current_config.401_page_pathname );
	//...
	
	401_page = malloc( fsize(current_config.401_page_pathname) );
	
}
*/

static socket_t _create_server_socket() // MEMO: se fallisce esce direttamente con "exit" (fatal error)
{
	int server_socket_fd;
		
	if( current_config.use_ipv6 == true )
	{
		// use IPv6 socket (support both IPv4 and IPv6 clients)
		server_socket_fd = socket( PF_INET6, SOCK_STREAM, 0 );
		if(-1 == server_socket_fd)
		{
			uhttps_log_error("[fatal error] unable to create TCP socket");
			exit(1);
		}
		
		struct sockaddr_in6 server_socket_address;	
		memset( &server_socket_address, 0, sizeof(server_socket_address) );
		server_socket_address.sin6_family = AF_INET6;
		server_socket_address.sin6_flowinfo = 0;
		server_socket_address.sin6_port = htons( current_config.listening_port );
		server_socket_address.sin6_addr = in6addr_any; // ascolto tutti gli indirizzi
		
		if( -1 ==
			bind( server_socket_fd, (const struct sockaddr*) &server_socket_address, sizeof(server_socket_address) )
		) {
			uhttps_log_error("[fatal error] unable to bind socket address. Please check your permissions.");
			exit(1);
		}
	}
	else
	{
		// use IPv4 socket
		server_socket_fd = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
		if(-1 == server_socket_fd)
		{
			uhttps_log_error("[fatal error] unable to create TCP socket");
			exit(1);
		}
		
		struct sockaddr_in server_socket_address;	
		memset( &server_socket_address, 0, sizeof(server_socket_address) );
		server_socket_address.sin_family = AF_INET;
		server_socket_address.sin_port = htons( current_config.listening_port );
		server_socket_address.sin_addr.s_addr = htonl(INADDR_ANY); // ascolto tutti gli indirizzi
		
		if( -1 ==
			bind( server_socket_fd, (const struct sockaddr*) &server_socket_address, sizeof(server_socket_address) )
		) {
			uhttps_log_error("[fatal error] unable to bind socket address. Please check your permissions.");
			exit(1);
		}
	}
	
	// TODO? imposta alcune opzioni per il listening socket (SO_REUSEADDR)
	
	// apre la porta di ascolto
	if( -1 ==
		listen( server_socket_fd, 10 ) // accetto max 10 connessioni pending
	  ) {
			uhttps_log_error("[fatal error] unable to open port %d", current_config.listening_port );
			exit(1);
		}
	  
	return server_socket_fd;
}

static void _close_server_socket()
{
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		shutdown( server_socket, 2 );
		closesocket( server_socket );
	#else
		shutdown( server_socket, SHUT_RDWR );
		close( server_socket );
	#endif
}

static socket_t _accept_a_connection( socket_t server_socket ) // MEMO: ritorna "-1" in caso di errore
{
	int client_connected_socket_fd =
		accept( server_socket, NULL, NULL );
	
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		if( client_connected_socket_fd == INVALID_SOCKET )
	#else
		if( client_connected_socket_fd == -1 )
	#endif
		{
			//uhttps_log_error("[warn] error while creating connection: %s", strerror(errno) );
			//strerror_r(...);
			return(-1);
		}
	// else
	return client_connected_socket_fd;
}

static void _parse_args( int argc, char *argv[] )
{
	if(argc==1) return;
	//else
	current_config_pathname = argv[1];
}

static char* _get_cwd() // ritorna la current working directory
{
	String r = calloc( _UHTTPS_MAX_PATHNAME_LENGHT, 1 );
	if(r==NULL) return("");
	
	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
		_getcwd( r, _UHTTPS_MAX_PATHNAME_LENGHT );
	#else
		getcwd( r, _UHTTPS_MAX_PATHNAME_LENGHT );
	#endif
	
	if(r==NULL)
		return("");
	else
		return(r);
}


int main( int argc, char *argv[] )
{
	uhttps_log_error("[notice] uhttps started, initializing...");
	uhttps_log_error("[notice] current working directory is \"%s\" all paths are relative to this", _get_cwd() );
		
	// parsa gli argomenti della linea di comando
	_parse_args( argc, argv );
	
	// inizializza i settaggi a default
	uhttps_configuration_init( &current_config );
	// carica tutto il file di configurazione in memoria e lo chiude
	uhttps_log_error("[notice] reading configuration from \"%s\"", current_config_pathname );
	uhttps_configuration_read( &current_config, current_config_pathname );
// DEBUG: stampa i settaggi del server
// DEBUG: uhttps_configuration_print( &current_config );
	
	// se il logfile NON è stdout, stacca il processo dalla console di invocazione
	if(( strcmp( current_config.log_pathname, "stdout" ) != 0 ) &&
	   ( strcmp( current_config.log_pathname, "stderr" ) != 0 ))
	{
		uhttps_log_error("[notice] main process is going to be demonized, this is the last output on this console.");
		_detach_main_process();
	}
	// se necessario apre il logfile in scrittura
	uhttps_logfile_open( current_config.log_pathname );
	atexit( uhttps_logfile_close ); // registra un exit handler per chiudere il logfile (fatto cmq automaticamente dal SO)

	// carica tutto il DB degli utenti in memoria e lo chiude
	uhttps_usersdb_read( &registered_usersdb, current_config.usersdb_pathname );
	//atexit(...) -> free(registered_usersdb); -> fatto automaticamente dal SO
	
	// inizializza la cache dei file (inizialmente vuota) secondo i settaggi
	uhttps_init_file_cache( &current_config );
	
	// registra i signal handlers
	_register_signal_handlers();
		
	// if using multithreading we need to initialize the mutexes
	if( uhttps_is_multithreaded( &current_config ) )
		_init_mutexes();

	#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	// inizializza la Winsock DLL
		WSADATA wsaData;
		if( WSAStartup( MAKEWORD (1, 1), &wsaData) != 0 )
		{
			uhttps_log_error("[fatal error] unable load Winsock DLL: %s", WSAGetLastError() );
			exit(1);
		}
		atexit( WSACleanup );
	#endif
	
	// crea il server socket per l'ascolto delle connessioni
	uhttps_log_error("[notice] starting listening connections on port %d...", current_config.listening_port);
	server_socket = _create_server_socket();
	atexit( _close_server_socket );
	
	server_is_opened = true;
	while( server_is_opened )
	{
		// attende la connessione di un client (può essere interrotta da un segnale)
		socket_t client_connected_socket =
			_accept_a_connection( server_socket );
		if( client_connected_socket != -1 ) // se la connessione è stata creata correttamente...
		{
			_increment_connected_clients_count(); // TODO? spostare in _client_handler()
		
			// avvia un processo/thread per la gestione della richiesta
			_client_handler( client_connected_socket );
		
			// se si è superato il limite di clients ammessi attende che uno termini
			while( _read_connected_clients_count() >= current_config.max_connected_clients_count )
			{
				#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
					Sleep(1000);
				#else
					sleep(1);
				#endif
			}
			// ALTERNATIVA: uso di Condition Variables -> PROBLEMA: NON supportate da Windows
		}
		
		// controlla se è arrivato un segnale che richiede la re-inizializzazione del server
		if( _current_config_reload_signal == true )
		{		
			// attende che tutti i threads secondari terminino
			// OSS.: nel frattempo non accetta altre connessioni e quindi non avvia altri threads...
			while( _read_connected_clients_count() > 0 )
			{
				#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
					Sleep(1000);
				#else
					sleep(1);
				#endif
			}
			
			// ora tutti i threads secondari sono terminati, si può modificare la struct current_config senza problemi di concorrenza...
			
			uhttps_log_error("[notice] re-loading configuration from \"%s\"", current_config_pathname );
			const short int LISTENING_PORT_OLD = current_config.listening_port;
			uhttps_configuration_read( &current_config, current_config_pathname );
			uhttps_logfile_close();
			uhttps_logfile_open( current_config.log_pathname );
			uhttps_usersdb_free( &registered_usersdb );
			uhttps_usersdb_read( &registered_usersdb, current_config.usersdb_pathname );
			// NO? free current file cache
			// NO? uhttps_init_file_cache( &current_config );
			if( current_config.listening_port != LISTENING_PORT_OLD ) // se la listening_port è cambiata, chiude e riapre la porta di ascolto
			{
				_close_server_socket();
				server_socket = _create_server_socket();
			}
			
			_current_config_reload_signal = false; // resetta il flag
		} // END IF
		
	} // END WHILE( server_is_opened )
	
	
	// wait for all child processes / secondary threads to terminate
	while( connected_clients_count!=0 )
	{
		#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
			Sleep(1000);
		#else
			sleep(1); // OSS.: può essere interrotto da SIGCHLD
		#endif
	}
	
	uhttps_log_error("[notice] last client disconnected, exiting server cleanly...");
	
	// exit cleanly
	exit(0);
}


